#include "Segment.h"

// STL

#include <iterator>
#include <sstream>

// Ice
#include <IceUtil/Time.h>

// Eigen
#include <Eigen/Core>
#include <Eigen/Geometry>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/TimeUtil.h>

#include <RobotAPI/libraries/core/FramedPose.h>

#include <RobotAPI/libraries/aron/common/aron_conversions.h>

#include <RobotAPI/libraries/armem/client/Writer.h>
#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/libraries/armem/core/Time.h>
#include <RobotAPI/libraries/armem/core/aron_conversions.h>
#include <RobotAPI/libraries/armem/core/workingmemory/visitor.h>
#include <RobotAPI/libraries/armem/server/MemoryToIceAdapter.h>

#include <RobotAPI/libraries/armem_robot/aron/Robot.aron.generated.h>
#include <RobotAPI/libraries/armem_robot/robot_conversions.h>

#include <RobotAPI/libraries/armem_robot_state/aron/Transform.aron.generated.h>
#include <RobotAPI/libraries/armem_robot_state/aron_conversions.h>
#include <RobotAPI/libraries/armem_robot_state/common/localization/TransformHelper.h>
#include <RobotAPI/libraries/armem_robot_state/common/localization/types.h>

namespace armarx::armem::server::robot_state::localization
{

    Segment::Segment(armem::server::MemoryToIceAdapter& memoryToIceAdapter,
                     std::mutex& memoryMutex) :
        iceMemory(memoryToIceAdapter), memoryMutex(memoryMutex)
    {
        Logging::setTag("LocalizationSegment");
    }

    Segment::~Segment() = default;

    void Segment::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(p.coreSegment,
                       prefix + "seg.localization.CoreSegment",
                       "Name of the object instance core segment.");
        defs->optional(p.maxHistorySize,
                       prefix + "seg.localization.MaxHistorySize",
                       "Maximal size of object poses history (-1 for infinite).");
    }

    void Segment::init()
    {
        ARMARX_CHECK_NOT_NULL(iceMemory.workingMemory);

        ARMARX_INFO << "Adding core segment '" << p.coreSegment << "'";

        coreSegment = &iceMemory.workingMemory->addCoreSegment(
                          p.coreSegment, arondto::Transform::toAronType());
        coreSegment->setMaxHistorySize(p.maxHistorySize);
    }

    void Segment::connect(viz::Client arviz)
    {
        // this->visu = std::make_unique<Visu>(arviz, *this);
    }

    std::unordered_map<std::string, std::vector<Eigen::Affine3f>>
            Segment::getRobotFramePoses(const armem::Time& timestamp) const
    {
        using common::robot_state::localization::TransformHelper;
        using common::robot_state::localization::TransformQuery;

        std::unordered_map<std::string, std::vector<Eigen::Affine3f>> frames;

        const auto& localizationCoreSegment =
            iceMemory.workingMemory->getCoreSegment(p.coreSegment);
        const auto knownRobots = simox::alg::get_keys(localizationCoreSegment);

        for (const auto& robotName : knownRobots)
        {

            TransformQuery query{.header{.parentFrame = GlobalFrame,
                                         .frame       = "root", // TODO, FIXME
                                         .agent       = robotName,
                                         .timestamp   = timestamp}};

            const auto result = TransformHelper::lookupTransformChain(localizationCoreSegment, query);

            if (not result)
            {
                // TODO
                continue;
            }

            frames.emplace(robotName, result.transforms);
        }

        ARMARX_INFO << deactivateSpam(10)
                    << "Number of known robot pose chains: " << frames.size();

        return frames;
    }

    std::unordered_map<std::string, Eigen::Affine3f>
    Segment::getRobotGlobalPoses(const armem::Time& timestamp) const
    {

        using common::robot_state::localization::TransformHelper;
        using common::robot_state::localization::TransformQuery;

        std::unordered_map<std::string, Eigen::Affine3f> robotGlobalPoses;

        const auto& localizationCoreSegment =
            iceMemory.workingMemory->getCoreSegment(p.coreSegment);
        const auto knownRobots = simox::alg::get_keys(localizationCoreSegment);

        for (const auto& robotName : knownRobots)
        {

            TransformQuery query{.header{.parentFrame = GlobalFrame,
                                         .frame       = "root", // TODO, FIXME
                                         .agent       = robotName,
                                         .timestamp   = timestamp}};

            const auto result = TransformHelper::lookupTransform(localizationCoreSegment, query);

            if (not result)
            {
                // TODO
                continue;
            }

            robotGlobalPoses.emplace(robotName, result.transform.transform);
        }

        ARMARX_INFO << deactivateSpam(10)
                    << "Number of known robot poses: " << robotGlobalPoses.size();

        return robotGlobalPoses;
    }

    bool Segment::storeTransform(const armarx::armem::robot_state::Transform& transform)
    {
        const auto& timestamp = transform.header.timestamp;

        const MemoryID providerID =
            coreSegment->id().withProviderSegmentName(transform.header.agent);
        if (not coreSegment->hasProviderSegment(providerID.providerSegmentName))
        {
            coreSegment->addProviderSegment(providerID.providerSegmentName, arondto::Transform::toAronType());
        }

        Commit commit;

        EntityUpdate& update = commit.updates.emplace_back();
        update.entityID =
            providerID.withEntityName(transform.header.parentFrame + "," + transform.header.frame);
        update.timeArrived = update.timeCreated = update.timeSent = timestamp;

        arondto::Transform aronTransform;
        toAron(aronTransform, transform);
        update.instancesData = {aronTransform.toAron()};

        const armem::CommitResult result = iceMemory.commit(commit);
        return result.allSuccess();
    }

    // std::unordered_map<armem::MemoryID, ::armarx::armem::articulated_object::ArticulatedObjectDescription> Segment::getKnownObjectClasses() const
    // {
    //     std::unordered_map<armem::MemoryID, ::armarx::armem::articulated_object::ArticulatedObjectDescription> objects;

    //     for (const auto& [_, provSeg] : iceMemory.workingMemory->getCoreSegment(p.coreClassSegmentName))
    //     {
    //         for (const auto& [name, entity] :  provSeg.entities())
    //         {
    //             const auto& entityInstance = entity.getLatestSnapshot().getInstance(0);
    //             const auto description = articulated_object::convertRobotDescription(entityInstance);

    //             if (not description)
    //             {
    //                 ARMARX_WARNING << "Could not convert entity instance to 'RobotDescription'";
    //                 continue;
    //             }

    //             ARMARX_INFO << "Key is " << armem::MemoryID(entity.id());

    //             objects.emplace(armem::MemoryID(entity.id()), *description);
    //         }
    //     }

    //     ARMARX_IMPORTANT << "Number of known articulated object classes: " << objects.size();

    //     return objects;
    // }

    // void Segment::RemoteGui::setup(const Segment& data)
    // {
    //     using namespace armarx::RemoteGui::Client;

    //     maxHistorySize.setValue(std::max(1, int(data.p.maxHistorySize)));
    //     maxHistorySize.setRange(1, 1e6);
    //     infiniteHistory.setValue(data.p.maxHistorySize == -1);
    //     discardSnapshotsWhileAttached.setValue(data.p.discardSnapshotsWhileAttached);

    //     GridLayout grid;
    //     int row = 0;
    //     grid.add(Label("Max History Size"), {row, 0}).add(maxHistorySize, {row, 1});
    //     row++;
    //     grid.add(Label("Infinite History Size"), {row, 0}).add(infiniteHistory, {row, 1});
    //     row++;
    //     grid.add(Label("Discard Snapshots while Attached"), {row, 0}).add(discardSnapshotsWhileAttached, {row, 1});
    //     row++;

    //     group.setLabel("Data");
    //     group.addChild(grid);
    // }

    // void Segment::RemoteGui::update(Segment& data)
    // {
    //     if (infiniteHistory.hasValueChanged() || maxHistorySize.hasValueChanged()
    //         || discardSnapshotsWhileAttached.hasValueChanged())
    //     {
    //         std::scoped_lock lock(data.memoryMutex);

    //         if (infiniteHistory.hasValueChanged() || maxHistorySize.hasValueChanged())
    //         {
    //             data.p.maxHistorySize = infiniteHistory.getValue() ? -1 : maxHistorySize.getValue();
    //             if (data.coreSegment)
    //             {
    //                 data.coreSegment->setMaxHistorySize(long(data.p.maxHistorySize));
    //             }
    //         }

    //         data.p.discardSnapshotsWhileAttached = discardSnapshotsWhileAttached.getValue();
    //     }
    // }

} // namespace armarx::armem::server::robot_state::localization
