/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NlohmannJSONReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/data/AllNavigators.h>



namespace armarx::aron::dataIO::reader
{
    NlohmannJSONReader::NlohmannJSONReader(const std::string& n) :
        NlohmannJSONReader(nlohmann::json::parse(n))
    {
    }

    NlohmannJSONReader::NlohmannJSONReader(const nlohmann::json& n) :
        input(n)
    {
        if (!n.is_object())
        {
            throw error::AronException("NlohmannJSONReader", "NlohmannJSONReader", "Allowed are only objects in reader");
        }

        if (n.at(io::Data::READER_WRITER_TYPE_SLUG) != io::Data::READER_WRITER_DICT_TYPENAME_SLUG)
        {
            throw error::StringNotValidException("NlohmannJSONReader", "readStartList", "The json is not of dict type!", n[io::Data::READER_WRITER_TYPE_SLUG], io::Data::READER_WRITER_DICT_TYPENAME_SLUG);
        }

        auto newToken = std::make_shared<NlohmannJSONReaderToken>(n);
        stack.push(newToken);
    }

    ReaderInterface::ReadStartDictReturnType NlohmannJSONReader::readStartDict()
    {
        auto lastToken = stack.top();
        if (!readFirstStartDict)
        {
            // we already added in constructor
            readFirstStartDict = true;
            auto current_json = lastToken->getElement();
            int c = lastToken->getElementChildrenSize();
            return {c, true};
        }

        auto current_json = lastToken->getNextElement();
        if (current_json.is_null())
        {
            lastToken->increaseCounter(); // no need to call readEndDict anymore
            return {0, false};
        }

        if (!current_json.is_object())
        {
            throw error::AronException("NlohmannJSONReader", "readStartDict", "The json is not an object. Every part of the data must be an object containing the type and the value.");
        }

        if (current_json.at(io::Data::READER_WRITER_TYPE_SLUG) != io::Data::READER_WRITER_DICT_TYPENAME_SLUG)
        {
            throw error::StringNotValidException("NlohmannJSONReader", "readStartList", "The json is not of dict type!", current_json.at(io::Data::READER_WRITER_TYPE_SLUG), io::Data::READER_WRITER_DICT_TYPENAME_SLUG);
        }

        auto newToken = std::make_shared<NlohmannJSONReaderToken>(current_json);
        int c = newToken->getElementChildrenSize();
        stack.push(newToken);
        return {c, true};
    }

    bool NlohmannJSONReader::readEndDict()
    {
        auto token = stack.top();
        token->assertType(data::Descriptor::eDict);

        if (token->finishedElement())
        {
            stack.pop();

            if (!stack.empty())
            {
                auto lastToken = stack.top();
                lastToken->increaseCounter();
            }

            return true;
        }
        return false;
    }

    ReaderInterface::ReadStartListReturnType NlohmannJSONReader::readStartList()
    {
        auto lastToken = stack.top();

        auto current_json = lastToken->getNextElement();
        if (current_json.is_null())
        {
            lastToken->increaseCounter(); // no need to call readEndList anymore
            return {0, false};
        }

        if (!current_json.is_object())
        {
            throw error::AronException("NlohmannJSONReader", "readStartList", "The json is not an object. Every part of the data must be an object containing the type and the value.");
        }

        if (current_json.at(io::Data::READER_WRITER_TYPE_SLUG) != io::Data::READER_WRITER_LIST_TYPENAME_SLUG)
        {
            throw error::StringNotValidException("NlohmannJSONReader", "readStartList", "The json is not of list type!", current_json.at(io::Data::READER_WRITER_TYPE_SLUG), io::Data::READER_WRITER_LIST_TYPENAME_SLUG);
        }

        auto newToken = std::make_shared<NlohmannJSONReaderToken>(current_json);
        int c = newToken->getElementChildrenSize();
        stack.push(newToken);
        return {c, true};
    }

    bool NlohmannJSONReader::readEndList()
    {
        auto token = stack.top();
        token->assertType(data::Descriptor::eList);

        if (token->finishedElement())
        {
            stack.pop();

            auto lastToken = stack.top();
            lastToken->increaseCounter();

            return true;
        }
        return false;
    }

    // Complex Data
    ReaderInterface::ReadStartNDArrayReturnType NlohmannJSONReader::readStartNDArray()
    {
        auto lastToken = stack.top();

        auto current_json = lastToken->getNextElement();
        if (current_json.is_null())
        {
            lastToken->increaseCounter();
            return {{}, "", false};
        }

        if (current_json.at(io::Data::READER_WRITER_TYPE_SLUG) != io::Data::READER_WRITER_NDARRAY_TYPENAME_SLUG)
        {
            throw error::StringNotValidException("NlohmannJSONReader", "readStartNDArray", "The json is not of ndarray type!", current_json.at(io::Data::READER_WRITER_TYPE_SLUG), io::Data::READER_WRITER_NDARRAY_TYPENAME_SLUG);
        }

        std::vector<int> dims = current_json.at(io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG);
        std::string type = current_json.at(io::Data::READER_WRITER_NDARRAY_TYPE_SLUG);
        return {dims, type, true};
    }

    void NlohmannJSONReader::readEndNDArray(unsigned char* data)
    {
        if (data == NULL)
        {
            throw error::AronException("NlohmannJSONReader", "readEndNDArray", "The corresponding data pointer of an complex aron ptr is NULL.");
        }

        auto lastToken = stack.top();

        auto current_json = lastToken->getNextElement();
        std::vector<int> dims = current_json.at(io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG);
        std::vector<unsigned char> d = current_json.at(io::Data::READER_WRITER_NDARRAY_DATA_SLUG);
        memcpy(data, d.data(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
        lastToken->increaseCounter();
    }

    // Read primitives
    ReaderInterface::ReadPrimitiveReturnType NlohmannJSONReader::readStartPrimitive()
    {
        auto lastToken = stack.top();

        auto current_json = lastToken->getNextElement();
        if (current_json.is_null())
        {
            lastToken->increaseCounter();
            return {false};
        }

        return {true};
    }

    void NlohmannJSONReader::readEndPrimitive(int& t)
    {
        auto lastToken = stack.top();
        auto j = lastToken->getNextElement();

        if (j.at(io::Data::READER_WRITER_TYPE_SLUG) != io::Data::READER_WRITER_INT_TYPENAME_SLUG)
        {
            throw error::StringNotValidException("NlohmannJSONReader", "readEndPrimitive", "The json is not of int type!", j.at(io::Data::READER_WRITER_TYPE_SLUG), io::Data::READER_WRITER_INT_TYPENAME_SLUG);
        }

        lastToken->increaseCounter();
        t = j.at(io::Data::READER_WRITER_VALUE_SLUG);
    }

    void NlohmannJSONReader::readEndPrimitive(long& t)
    {
        auto lastToken = stack.top();
        auto j = lastToken->getNextElement();

        if (j.at(io::Data::READER_WRITER_TYPE_SLUG) != io::Data::READER_WRITER_LONG_TYPENAME_SLUG)
        {
            throw error::StringNotValidException("NlohmannJSONReader", "readEndPrimitive", "The json is not of long type!", j.at(io::Data::READER_WRITER_TYPE_SLUG), io::Data::READER_WRITER_LONG_TYPENAME_SLUG);
        }

        lastToken->increaseCounter();
        t = j.at(io::Data::READER_WRITER_VALUE_SLUG);
    }

    void NlohmannJSONReader::readEndPrimitive(float& t)
    {
        auto lastToken = stack.top();
        auto j = lastToken->getNextElement();

        if (j.at(io::Data::READER_WRITER_TYPE_SLUG) != io::Data::READER_WRITER_FLOAT_TYPENAME_SLUG)
        {
            throw error::StringNotValidException("NlohmannJSONReader", "readEndPrimitive", "The json is not of float type!", j.at(io::Data::READER_WRITER_TYPE_SLUG), io::Data::READER_WRITER_FLOAT_TYPENAME_SLUG);
        }

        lastToken->increaseCounter();
        t = j.at(io::Data::READER_WRITER_VALUE_SLUG);
    }

    void NlohmannJSONReader::readEndPrimitive(double& t)
    {
        auto lastToken = stack.top();
        auto j = lastToken->getNextElement();

        if (j.at(io::Data::READER_WRITER_TYPE_SLUG) != io::Data::READER_WRITER_DOUBLE_TYPENAME_SLUG)
        {
            throw error::StringNotValidException("NlohmannJSONReader", "readEndPrimitive", "The json is not of double type!", j.at(io::Data::READER_WRITER_TYPE_SLUG), io::Data::READER_WRITER_DOUBLE_TYPENAME_SLUG);
        }

        lastToken->increaseCounter();
        t = j.at(io::Data::READER_WRITER_VALUE_SLUG);
    }

    void NlohmannJSONReader::readEndPrimitive(std::string& t)
    {
        auto lastToken = stack.top();
        auto j = lastToken->getNextElement();

        if (j.at(io::Data::READER_WRITER_TYPE_SLUG) != io::Data::READER_WRITER_STRING_TYPENAME_SLUG)
        {
            throw error::StringNotValidException("NlohmannJSONReader", "readEndPrimitive", "The json is not of string type!", j.at(io::Data::READER_WRITER_TYPE_SLUG), io::Data::READER_WRITER_STRING_TYPENAME_SLUG);
        }

        lastToken->increaseCounter();
        t = j.at(io::Data::READER_WRITER_VALUE_SLUG);
    }

    void NlohmannJSONReader::readEndPrimitive(bool& t)
    {
        auto lastToken = stack.top();
        auto j = lastToken->getNextElement();

        if (j.at(io::Data::READER_WRITER_TYPE_SLUG) != io::Data::READER_WRITER_BOOL_TYPENAME_SLUG)
        {
            throw error::StringNotValidException("NlohmannJSONReader", "readEndPrimitive", "The json is not of bool type!", j.at(io::Data::READER_WRITER_TYPE_SLUG), io::Data::READER_WRITER_BOOL_TYPENAME_SLUG);
        }

        lastToken->increaseCounter();
        t = j.at(io::Data::READER_WRITER_VALUE_SLUG);
    }

    std::string NlohmannJSONReader::readKey()
    {
        auto token = stack.top();
        return token->getCurrentKey();
    }

    void NlohmannJSONReader::loadMember(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    void NlohmannJSONReader::readNull()
    {
        auto token = stack.top();
        auto current_json = token->getNextElement();
        if (!current_json.is_null())
        {
            throw error::AronException("NlohmannJSONReader", "readNull", "The value of a json is not null!");
        }
        token->increaseCounter();
    }

    // Helper functions
    data::Descriptor NlohmannJSONReader::getTypeOfNext() const
    {
        if (stack.empty())
        {
            return data::Descriptor::eDict;
        }
        auto token = stack.top();
        return token->getTypeOfNext();
    }
}
