/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "Converter.h"

namespace armarx::aron::typeIO
{
    void Converter::ReadAndConvert(ReaderInterface& reader, WriterInterface& writer)
    {
        type::Descriptor desc = reader.getTypeOfNext();
        switch (desc)
        {
            case type::Descriptor::eObject:
            {
                const auto readStartObject = reader.readStartObject();
                writer.writeStartObject({readStartObject.name, readStartObject.maybe});

                for (unsigned int i = 0; i < readStartObject.elements; ++i)
                {
                    std::string key = reader.readKey();
                    writer.writeKey(key);
                    Converter::ReadAndConvert(reader, writer);
                }

                writer.writeEndObject();
                reader.readEndObject();
                break;
            }
            case type::Descriptor::eDict:
            {
                const auto readStartDict = reader.readStartDict();
                writer.writeStartDict({readStartDict.maybe});

                Converter::ReadAndConvert(reader, writer);

                writer.writeEndDict();
                reader.readEndDict();
                break;
            }
            case type::Descriptor::eTuple:
            {
                const auto readStartTuple = reader.readStartTuple();
                writer.writeStartTuple({readStartTuple.maybe});

                for (unsigned int i = 0; i < readStartTuple.elements; ++i)
                {
                    Converter::ReadAndConvert(reader, writer);
                }

                writer.writeEndTuple();
                reader.readEndTuple();
                break;
            }
            case type::Descriptor::eList:
            {
                const auto readStartList = reader.readStartList();
                writer.writeStartList({readStartList.maybe});

                Converter::ReadAndConvert(reader, writer);

                writer.writeEndList();
                reader.readEndList();
                break;
            }
            case type::Descriptor::eEigenMatrix:
            {
                const auto readEigenMatrix = reader.readEigenMatrix();
                writer.writeEigenMatrix({readEigenMatrix.rows, readEigenMatrix.cols, readEigenMatrix.type, readEigenMatrix.maybe});
                break;
            }
            case type::Descriptor::eEigenQuaternion:
            {
                const auto readEigenQuaternion = reader.readEigenQuaternion();
                writer.writeEigenQuaternion({readEigenQuaternion.type, readEigenQuaternion.maybe});
                break;
            }
            case type::Descriptor::eIVTCByteImage:
            {
                const auto readIVTCByteImage = reader.readIVTCByteImage();
                writer.writeIVTCByteImage({readIVTCByteImage.maybe});
                break;
            }
            case type::Descriptor::eOpenCVMat:
            {
                const auto readOpenCVMat = reader.readOpenCVMat();
                writer.writeOpenCVMat({readOpenCVMat.maybe});
                break;
            }
            case type::Descriptor::ePCLPointCloud:
            {
                const auto readPCLPointCloud = reader.readPCLPointCloud();
                writer.writePCLPointCloud({readPCLPointCloud.type, readPCLPointCloud.maybe});
                break;
            }
            case type::Descriptor::ePosition:
            {
                const auto readPosition = reader.readPosition();
                writer.writePosition({readPosition.maybe});
                break;
            }
            case type::Descriptor::eOrientation:
            {
                const auto readOrientation = reader.readOrientation();
                writer.writeOrientation({readOrientation.maybe});
                break;
            }
            case type::Descriptor::ePose:
            {
                const auto readPose = reader.readPose();
                writer.writePose({readPose.maybe});
                break;
            }
            case type::Descriptor::eInt:
            {
                const auto readInt = reader.readInt();
                writer.writeInt({readInt.maybe});
                break;
            }
            case type::Descriptor::eLong:
            {
                const auto readLong = reader.readLong();
                writer.writeLong({readLong.maybe});
                break;
            }
            case type::Descriptor::eFloat:
            {
                const auto readFloat = reader.readFloat();
                writer.writeFloat({readFloat.maybe});
                break;
            }
            case type::Descriptor::eDouble:
            {
                const auto readDouble = reader.readDouble();
                writer.writeDouble({readDouble.maybe});
                break;
            }
            case type::Descriptor::eString:
            {
                const auto readString = reader.readString();
                writer.writeString({readString.maybe});
                break;
            }
            case type::Descriptor::eBool:
            {
                const auto readBool = reader.readBool();
                writer.writeBool({readBool.maybe});
                break;
            }
            case type::Descriptor::eTime:
            {
                const auto readTime = reader.readTime();
                writer.writeTime({readTime.maybe});
                break;
            }
            default:
            {
                throw error::DescriptorNotValidException("LegacyAronTypeReader", "ConvertFromReader", "Type could not be resolved", desc);
            }
        }
    }
}
