/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>
#include <vector>

// ArmarX
#include <RobotAPI/libraries/aron/core/Config.h>
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/io/Data.h>

namespace armarx::aron::typeIO
{
    template<typename ElementTypename>
    class ReaderTokenInterface;

    template<typename ElementTypename>
    using ReaderTokenInterfacePtr = std::shared_ptr<ReaderTokenInterface<ElementTypename>>;

    template<typename ElementTypename>
    class ReaderTokenInterface
    {
    public:
        using PointerType = ReaderTokenInterfacePtr<ElementTypename>;

    public:
        // Interface
        virtual type::Descriptor getTypeOfNext() const = 0;
        virtual ElementTypename getNextElement() const = 0;

        // General implementation
        std::string getElementName() const
        {
            switch (descriptor)
            {
                case type::Descriptor::eObject:
                case type::Descriptor::eIntEnum:
                {
                    return elementName;
                }
                default:
                    throw error::DescriptorNotValidException("ReaderTokenInterface", "getElementName", "Could not get object nme of a non-object or non-enum token", descriptor);
            }
        }

        type::Descriptor getDescriptor() const
        {
            return descriptor;
        }

        unsigned int getElementChildrenSize() const
        {
            return childrenSize;
        }

        bool finishedElement() const
        {
            return currentIndex >= getElementChildrenSize();
        }

        std::string getCurrentKey() const
        {
            assertIndexInBounds();

            // ugly workaround
            if (this->currentKey == "") // probably in a dict where the caller does not know which keys are there
            {
                if (this->allMemberNames.size() == 0)
                {
                    throw error::AronException("typeIO::ReaderTokenInterface", "getCurrentKey", "The currentKey is not set and the list of all keys is empty!");
                }
                return this->allMemberNames[this->currentIndex];
            }
            return this->currentKey;
        }

        void setCurrentKey(const std::string& s)
        {
            currentKey = s;
        }

        void assertType(type::Descriptor t) const
        {
            if (getDescriptor() != t)
            {
                throw error::DescriptorNotValidException("typeIO::ReaderTokenInterface", "assertType", "The type was not equal.", t);
            }
        }

        void assertIndexInBounds() const
        {
            if (this->currentIndex >= getElementChildrenSize())
            {
                throw error::SizeNotValidException("typeIO::ReaderTokenInterface", "assertIndexInBounds", "Index out of bounds.", this->currentIndex, getElementChildrenSize());
            }
        }

        ElementTypename getNextElementAndIncreaseCnt()
        {
            ElementTypename next = getNextElement();
            this->currentIndex++;
            return next;
        }

    protected:
        // members
        type::Descriptor descriptor = type::Descriptor::eUnknown;
        ElementTypename element;
        std::string elementName = "";

        unsigned int childrenSize = 0;
        unsigned int currentIndex = 0;

        std::string currentKey = "";
        std::vector<std::string> allMemberNames = {};
    };
}
